from typing import Union

import numpy as np
import pandas as pd
import requests

from .features import Question


class KoboForm:
    def __init__(self, uid: str) -> None:
        self.uid = uid
        self.headers = None
        self.metadata = {}
        self.data = None
        self.has_geo = False
        self.geo = []
        self.has_repeats = False
        self.repeats = {}
        self.__root_structure = []
        self.__repeats_structure = {}
        self.__asset = None
        self.__content = None
        self.__columns_as = "name"
        self.__choices_as = "name"
        self.naming_conflicts = None
        self.separator = "|"
        self.__initial_separator = " "

    def __repr__(self):
        return f"KoboForm('{self.uid}')"

    def fetch_data(self) -> Union[pd.DataFrame, dict]:
        """Fetch the form's data and store them as a Pandas DF in the attribute `data`.
        If the form has repeat groups, extract them as separate DFs"""

        self._get_survey()

        # It's possible for a form to have no "choices" (corresponds to
        # a XLSForm without a tab "choices"). In this case we don't call
        # the method '_get_choices'
        if "choices" in self.__content:
            self._get_choices()

        # Fetch the data
        res = requests.get(url=self.__url_data, headers=self.headers)

        # If error while fetching the data, return an empty DF
        if res.status_code != 200:
            return pd.DataFrame()

        data = res.json()["results"]

        self.data = pd.DataFrame(data)

        self._remove_unused_columns()

        # For columns containing the group(s) they belong to in their name, remove it to only
        # keep the name of the column
        __dict_rename = {}
        for c in self.data.columns:
            __dict_rename[c] = c.split("/")[-1]

        self.data.rename(columns=__dict_rename, inplace=True)

        # Add a column '_index' that can be used to join the parent DF
        # with the children DFs (which have the column '_parent_index')
        self.data["_index"] = self.data.index + 1

        # If the form has at least one repeat group
        if self.has_repeats:

            self._extract_repeats(data)

            # In the parent DF delete the columns that contain the repeat groups
            # In the API there is a column with the same name as the name of
            # the repeat group + the suffix '_count' just before the repeat group.
            # We can delete it
            repeats_count = [f"{c}_count" for c in self.repeats.keys()]
            to_delete = list(self.repeats.keys()) + repeats_count

            self.data.drop(columns=to_delete, inplace=True)

        # The JSON object returned by the API containing the form data doesn't
        # have properties for empyty columns. So, here all empty columns are missing.
        # We need to add them
        for q in self.__root_structure:
            if q.name not in self.data.columns:
                self.data[q.name] = np.nan
        if self.has_repeats:
            for k, v in self.repeats.items():
                for q in self.__repeats_structure[k]["columns"]:
                    if q.name not in v.columns:
                        v[q.name] = np.nan

        if self.has_geo:
            self._split_gps_coords()

        # At this point we don't add or delete columns any more
        # so we can reorder the columns as they are in the API

        # the columns that are in the DF but not in the structure
        # will be moved to the end
        last_columns = [
            c
            for c in self.data.columns
            if c not in [q.name for q in self.__root_structure]
        ]

        columns_ordered = [q.name for q in self.__root_structure] + last_columns

        self.data = self.data[columns_ordered]

        if self.has_repeats:
            for k, v in self.repeats.items():
                columns_ordered = [
                    q.name for q in self.__repeats_structure[k]["columns"]
                ]

                # The column'_parent_index' will be in the last position
                columns_ordered.append("_parent_index")

                self.repeats[k] = self.repeats[k][columns_ordered]

        # We need to run `_change_choices` here in order to format the multiple choices
        # so that it's possible to go back and forth between name and label for the choices
        # In the Kobo API, multiple choices are seprated by ' '. We replace ' ' with self.separator
        self._change_choices(self.data, self.__root_structure, self.__choices_as)
        if self.has_repeats:
            for k, v in self.repeats.items():
                self._change_choices(
                    v, self.__repeats_structure[k]["columns"], self.__choices_as
                )
        self.__initial_separator = self.separator

    def display(self, columns_as: str = "name", choices_as: str = "name") -> None:
        """Update the DatFrames containing the data by using names or labels for
        the columns and/or the choices based on the values of the parameters `columns_as`
        and `choices_as`."""
        if columns_as not in ["name", "label"]:
            raise ValueError(
                f"'{columns_as}' is not an accepted value for the parameter 'columns_as'. Accepted values are 'name' or 'label'."
            )

        if choices_as not in ["name", "label"]:
            raise ValueError(
                f"'{choices_as}' is not an accepted value for the parameter 'choices_as'. Accepted values are 'name' or 'label'."
            )

        # Change the columns names
        if self.__columns_as != columns_as:
            self._rename_columns(self.__columns_as, columns_as)
            self.__columns_as = columns_as

        # Change the choices names
        if self.__choices_as != choices_as:
            self._change_choices(self.data, self.__root_structure, choices_as)
            if self.has_repeats:
                for k, v in self.repeats.items():
                    self._change_choices(
                        v, self.__repeats_structure[k]["columns"], choices_as
                    )
            self.__choices_as = choices_as

    def _get_survey(self) -> None:
        """Go through all the elements of the survey and build the root structure (and the structure
        of the repeat groups if any) as a list of `Question` objects. Each `Question` object has a name
        and a label so it's possible to display the data using any of the two."""

        if not self.__asset:
            self._fetch_asset()

        if "naming_conflicts" in self.__asset["summary"]:
            self.naming_conflicts = self.__asset["summary"]["naming_conflicts"]

        survey = self.__content["survey"]

        group_name = None
        group_label = None
        repeat_name = None
        repeat_label = None
        in_repeat = False

        for field in survey:

            # Identify groups and repeats if any
            if field["type"] == "begin_group":

                group_name = field["name"]
                if "label" in field:
                    group_label = field["label"]

            if field["type"] == "begin_repeat":

                repeat_name = field["name"]
                if "label" in field:
                    repeat_label = field["label"]

                in_repeat = True
                self.has_repeats = True
                self.__repeats_structure[repeat_name] = {}
                self.__repeats_structure[repeat_name]["columns"] = []
                self.__repeats_structure[repeat_name]["has_geo"] = False
                self.__repeats_structure[repeat_name]["geo"] = False

            if field["type"] == "end_group":
                group_name = None
                group_label = None

            if field["type"] == "end_repeat":
                in_repeat = False
                repeat_name = None
                repeat_label = None

            if (
                field["type"] != "begin_group"
                and field["type"] != "begin_repeat"
                and field["type"] != "end_group"
                and field["type"] != "end_repeat"
            ):

                name_q = field["name"]
                if "label" in field:
                    label_q = field["label"][0]
                else:
                    label_q = name_q

                if (
                    self.naming_conflicts
                    and name_q in self.naming_conflicts
                    and field["type"]
                    not in [
                        "start",
                        "end",
                        "today",
                        "username",
                        "deviceid",
                        "phonenumber",
                        "calculate",
                    ]
                ):
                    name_q = f"{name_q}_001"

                q = Question(name_q, field["type"], label_q)

                if field["type"] == "select_one" or field["type"] == "select_multiple":
                    q.select_from_list_name = field["select_from_list_name"]

                q.group_name = group_name
                q.group_label = group_label

                q.repeat_name = repeat_name
                q.repeat_label = repeat_label

                if in_repeat:
                    self.__repeats_structure[repeat_name]["columns"].append(q)
                    # Identify the geopoint if any
                    if field["type"] == "geopoint":
                        self.__repeats_structure[repeat_name]["has_geo"] = True
                        self.__repeats_structure[repeat_name]["geo"].append(q)
                else:
                    self.__root_structure.append(q)
                    # Identify the geopoint if any
                    if field["type"] == "geopoint":
                        self.has_geo = True
                        self.geo.append(q)

        self._rename_columns_labels_duplicates(self.__root_structure)
        if self.has_repeats:
            for k, repeat in self.__repeats_structure.items():
                self._rename_columns_labels_duplicates(repeat["columns"])

    def _get_choices(self):
        """For all the questions of type 'select_one' or 'select_multiple' assign their corresponding choices.
        Each choice has a name and label so it's possible to display the data using any of the two."""

        formatted_choices = {}
        choices = self.__content["choices"]
        for choice in choices:
            if choice["list_name"] not in formatted_choices:
                formatted_choices[choice["list_name"]] = []
            formatted_choices[choice["list_name"]].append(
                {"name": choice["name"], "label": choice["label"][0]}
            )

        for q in self.__root_structure:
            if q.type == "select_one" or q.type == "select_multiple":
                q.choices = formatted_choices[q.select_from_list_name]

        if self.has_repeats:
            for k, repeat in self.__repeats_structure.items():
                for q in repeat["columns"]:
                    if q.type == "select_one" or q.type == "select_multiple":
                        q.choices = formatted_choices[q.select_from_list_name]

    def _change_choices(
        self, df: pd.DataFrame, structure: list, choices_as: str
    ) -> None:
        """Change the choices for the columns of type 'select_one' and 'select_multiple'
        from name to label and vice versa."""
        for q in structure:
            if q.type == "select_one":
                column = getattr(q, self.__columns_as)
                for choice in q.choices:
                    df.loc[df[column] == choice[self.__choices_as], column] = choice[
                        choices_as
                    ]

            # Multiple choices
            if q.type == "select_multiple":
                column = getattr(q, self.__columns_as)
                unique_values = list(df[column].unique())

                for unique in unique_values:
                    if pd.isna(unique):
                        pass
                    else:
                        combinations = unique.split(self.__initial_separator)
                        new_choices = [
                            c[choices_as]
                            for c in q.choices
                            if c[self.__choices_as] in combinations
                        ]
                        new_choices_formatted = f"{self.separator}".join(new_choices)

                        df.loc[df[column] == unique, column] = new_choices_formatted

    def _fetch_asset(self):
        res = requests.get(url=self.__url_asset, headers=self.headers)
        self.__asset = res.json()
        self.__content = res.json()["content"]

    def _extract_from_asset(self, asset: dict) -> None:
        self.metadata["name"] = asset["name"]
        self.metadata["owner"] = asset["owner__username"]
        self.metadata["date_created"] = asset["date_created"]
        self.metadata["date_modified"] = asset["date_modified"]
        self.metadata["version_id"] = asset["version_id"]
        self.metadata["has_deployment"] = asset["has_deployment"]
        self.metadata["geo"] = asset["summary"]["geo"]

        self.__url_asset = asset["url"]
        self.__url_data = asset["data"]
        self.__base_url = "/".join(asset["downloads"][0]["url"].split("/")[:-1])

    def _rename_columns(self, old, new):
        """Used to change the columns names from name to label and vice versa"""
        dict_rename = {}
        for q in self.__root_structure:
            dict_rename[getattr(q, old)] = getattr(q, new)

        self.data.rename(columns=dict_rename, inplace=True)

        if self.has_repeats:
            for k, repeat in self.__repeats_structure.items():
                dict_rename = {}
                for q in repeat["columns"]:
                    dict_rename[getattr(q, old)] = getattr(q, new)
                self.repeats[k].rename(columns=dict_rename, inplace=True)

    def _extract_repeats(self, rows: list) -> None:
        """Extract all the questions part of repeat groups into separate DFs
        '_parent_index' is the column name used in Kobo in the child table
        when downloading the data, that allows to join the data with the parent table
        """
        repeats = {}
        for idx_parent, row in enumerate(rows):
            for column, value in row.items():
                if not column.startswith("_") and type(value) == list:
                    repeat_name = column.split("/")[-1]
                    if repeat_name not in repeats:
                        repeats[repeat_name] = []
                    for child in value:
                        child["_parent_index"] = idx_parent + 1
                    repeats[repeat_name] = repeats[repeat_name] + value

        for repeat_name, repeat_data in repeats.items():
            repeats[repeat_name] = pd.DataFrame(repeat_data)

            # If columns have a prefix composed of all their groups,
            # remove them
            dict_rename = {}
            for c in repeats[repeat_name].columns:
                dict_rename[c] = c.split("/")[-1]
            repeats[repeat_name].rename(columns=dict_rename, inplace=True)

            self.has_repeats = True
            self.repeats = repeats

    def _remove_unused_columns(self) -> None:
        """Remove the columns in the list `columns` if they are in the
        main `self.data` (before extracting the repeats)"""

        columns = [
            "_version_",
            "formhub/uuid",
            "meta/instanceID",
            "_xform_id_string",
            "_attachments",
            "meta/deprecatedID",
            "_geolocation",
        ]

        # We only try to delete the columns that are in the DataFrame
        to_delete = [c for c in columns if c in self.data.columns]

        if len(to_delete) > 0:
            self.data.drop(to_delete, axis=1, inplace=True)

    def _rename_columns_labels_duplicates(self, structure: list) -> None:
        """Identify the duplicates among the labels of all columns in ``structure`.
        In case of duplicates, rename the label by appending (x) at the end of the label.
        'x' being the number of time the duplicate has been encountered
        while going through all the columns."""
        all_labels = [q.label for q in structure]
        duplicate_labels = [l for l in all_labels if all_labels.count(l) > 1]
        duplicates_count = {}
        for d in duplicate_labels:
            duplicates_count[d] = 0

        for q in structure:
            if q.label in duplicates_count:
                duplicates_count[q.label] += 1
                q.label = f"{q.label} ({duplicates_count[q.label]})"

    def _split_gps_coords(self) -> None:
        """Split the columns of type 'geopoint' into 4 new columns
        'latitude', 'longitude', 'altitude', 'gps_precision'
        """

        base_geo_columns = ["latitude", "longitude", "altitude", "precision"]

        for g in self.geo:
            index_geo = self.__root_structure.index(g)
            new_geo_names = [f"_{g.name}_{c}" for c in base_geo_columns]
            new_geo_labels = [f"_{g.label}_{c}" for c in base_geo_columns]

            for idx, c in enumerate(new_geo_names):
                q = Question(c, "geo", new_geo_labels[idx])
                self.__root_structure.insert(index_geo + idx + 1, q)

            self.data[new_geo_names] = self.data[g.name].str.split(" ", expand=True)

        # XXX This has not been tested
        if self.has_repeats:
            for repeat_name, repeat in self.__repeats_structure.items():
                if repeat["has_geo"]:
                    for g in repeat["geo"]:
                        index_geo = repeat["columns"].index(g)
                        new_geo_names = [f"_{g.name}_{c}" for c in base_geo_columns]
                        new_geo_labels = [f"_{g.label}_{c}" for c in base_geo_columns]

                        for idx, c in enumerate(new_geo_names):
                            q = Question(c, "geo", new_geo_labels[idx])
                            repeat["columns"].append(index_geo + idx + 1, q)

                        self.data[new_geo_names] = self.repeats[repeat_name][
                            g.name
                        ].str.split(" ", expand=True)

    def download_form(self, format: str) -> None:
        """Given the uid of a form and a format ('xls' or 'xml')
        download the form in that format in the current directory"""

        if format not in ["xls", "xml"]:
            raise ValueError(
                f"The file format '{format}' is not supported. Recognized formats are 'xls' and 'xml'"
            )

        URL = f"{self.__base_url}/{self.uid}.{format}"
        filename = URL.split("/")[-1]

        r = requests.get(URL, headers=self.headers)

        with open(filename, "wb") as f:
            f.write(r.content)



class Question:
    def __init__(self, name: str, type: str, label: str = None) -> None:
        self.name = name
        self.type = type
        self.label = label
        self.group_name = None
        self.group_label = None
        self.repeat_name = None
        self.repeat_label = None
        self.select_from_list_name = None
        self.choices = None

    def __repr__(self):
        return f"Question('{self.name}, {self.type}, {self.label}')"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pykobo']

package_data = \
{'': ['*']}

install_requires = \
['pandas>=1.5.1,<2.0.0', 'requests>=2.28.1,<3.0.0']

setup_kwargs = {
    'name': 'pykobo',
    'version': '0.1.2',
    'description': 'A Python module to fetch data from the Kobo API',
    'long_description': '![PyPI](https://img.shields.io/pypi/v/pykobo)\n\n# pykobo\n\nPykobo is a Python module that fetches data from [Kobo](https://www.kobotoolbox.org/) forms via the [Kobo API](https://support.kobotoolbox.org/api.html).\n\nThe data is returned as a [pandas](https://pandas.pydata.org/) DataFrame making it easy to integrate in your workflow for further cleaning, filtering, aggregatation, etc...  \n\n## Functionalities\nPykobo can:\n* Get the list of forms a user has access to \n* Fetch the data of a form (with names or labels for the columns and choices) as a pandas DataFrame\n* Download a form in XLS or XML format\n\n## Install\n\n```bash\n$ pip install pykobo\n```\n\n## Examples\n\n### How to start\n\nNB: The example below uses [kf.kobotoolbox.org](https://kf.kobotoolbox.org/) but pykobo also works with [kobo.humanitarianresponse.info](https://kobo.humanitarianresponse.info/) and any other Kobo server.\n\nTo get your API token, see [here](https://support.kobotoolbox.org/api.html#getting-your-api-token).\n\n```python\nimport pykobo\n\n\nURL_KOBO_API = "https://kf.kobotoolbox.org/api/v2"\nMYTOKEN = "2bc8e0201d23dac4ec1c334107698147b81513a2"\n\n# Initialize the Manager object\nkm = pykobo.Manager(url_api=URL_KOBO_API, token=MYTOKEN)\n```\n\n### Get the list of forms you have access to\n\n```python\nmy_forms = km.get_forms()\n\n```\nThis returns a Python list of KoboForm objetcs\n\n```python\nprint(my_forms)\n\n[   KoboForm(\'tpz2buHAdXxcN0JVrZaSdk\'),\n    KoboForm(\'vyARFbyE8Gv3RUvXNfdTRj\'),\n    KoboForm(\'wogyYJzUu2ZFVnzqGg8K7q\'),\n    KoboForm(\'bQLZapErE3UqqG9Avntkhd\')]\n\n# Each form contains metadata\nfor f in my_forms:\n      print(f.metadata)\n\n{   \'date_created\': \'2022-07-14T20:44:11.929901Z\',\n    \'date_modified\': \'2022-10-02T07:49:19.714891Z\',\n    \'geo\': True,\n    \'has_deployment\': True,\n    \'name\': \'Household survey\',\n    \'owner\': \'pvernier\',\n    \'version_id\': \'aqUMoSqANiEgH3j4Nn3Cr7\'}\n{   \'date_created\': \'2022-07-14T12:41:14.665314Z\',\n    \'date_modified\': \'2022-09-28T11:55:15.408542Z\',\n    \'geo\': True,\n    \'has_deployment\': True,\n    \'name\': \'Health facilities monitoring\',\n    \'owner\': \'pvernier\',\n    \'version_id\': \'abLugnJGURSyyZ8RZxC0wQ\'}\n{   \'date_created\': \'2022-07-14T13:40:32.033446Z\',\n    \'date_modified\': \'2022-09-28T09:19:20.691620Z\',\n    \'geo\': False,\n    \'has_deployment\': True,\n    \'name\': \'Post emergency evaluation\',\n    \'owner\': \'pvernier\',\n    \'version_id\': \'aQQUmPns7xLUL4Ro0amqwS\'}\n{   \'date_created\': \'2022-09-14T16:54:06.990672Z\',\n    \'date_modified\': \'2022-09-20T13:27:52.410261Z\',\n    \'geo\': True,\n    \'has_deployment\': True,\n    \'name\': \'Identification of burnt areas\',\n    \'owner\': \'pvernier\',\n    \'version_id\': \'xes8JkQRpbDcbct9sqmCYZ\'}  \n```\n\n### Fetch a single form with its uid.\n\n```python\nuid = \'tpz2buHAdXxcN0JVrZaSdk\'\n\nmy_form = km.get_form(uid)\n\nprint(my_form.metadata)\n\n{   \'date_created\': \'2022-07-14T20:44:11.929901Z\',\n    \'date_modified\': \'2022-10-02T07:49:19.714891Z\',\n    \'geo\': True,\n    \'has_deployment\': True,\n    \'name\': \'Household survey\',\n    \'owner\': \'pvernier\',\n    \'version_id\': \'aqUMoSqANiEgH3j4Nn3Cr7\'}\n\n```\n\n\n### Fetch the data of a form\n\n```python\nmy_form.fetch_data()\n\n# The data is accessible via the `data` attribute as a pandas DataFrame\n\nprint(my_form.data)\n                             start                            end       today         username                  deviceid            phonenumber  ...     _submission_time _tags _notes _validation_status    _submitted_by _index\n0    2022-09-01T15:47:55.797+02:00  2022-09-01T15:51:48.302+02:00  2022-09-01  surveyer_1  collect:4vUec4gLVJx3GP1D                    NaN  ...  2022-09-01T13:52:04    []     []                 {}  surveyer_1      1\n1    2022-09-01T15:58:08.251+02:00  2022-09-01T16:08:14.548+02:00  2022-09-01  surveyer_1  collect:Xk9Z5f1VTW5nig68                    NaN  ...  2022-09-01T14:08:46    []     []                 {}  surveyer_1      2\n2    2022-09-01T14:05:08.484+02:00  2022-09-01T16:17:59.305+02:00  2022-09-01  surveyer_1  collect:0Y8Cozz5fzI8jczs                    NaN  ...  2022-09-01T14:18:36    []     []                 {}  surveyer_1      3\n3    2022-09-01T16:20:39.699+02:00  2022-09-01T16:32:03.393+02:00  2022-09-01  surveyer_1  collect:MPi52tvGiPY6AuK3                    NaN  ...  2022-09-01T14:32:27    []     []                 {}  surveyer_1      4\n...\n...\n[595 rows x 38 columns]\n\nprint(type(my_form.data))\n\n<class \'pandas.core.frame.DataFrame\'>\n\n# The method `fetch_data` returns the data using the Kobo columns and choices names\n\nprint(my_form.data.columns)\n\nIndex([\'start\', \'end\', \'today\', \'username\', \'deviceid\', \'phonenumber\', \'date\',\n       \'health_area\', \'village_name\', \'team_number\', \'cluster_number\',\n       \'household_number\', \'gps\', \'_gps_latitude\', \'_gps_longitude\',\n       \'_gps_altitude\', \'_gps_precision\', \'hhh_present\',\n       \'age_hhh\', \'consent\', \'number_children\',\n       \'__version__\', \'_id\', \'_uuid\', \'_status\', \'_submission_time\',\n       \'_tags\', \'_notes\', \'_validation_status\', \'_submitted_by\', \'_index\'],\n      dtype=\'object\')\n```\n\n### Display the data using Kobo labels for columns and/or choices\n\n\n```python\nmy_form.display(columns_as=\'label\', choices_as=\'label\')\n\nprint(my_form.data.columns)\n\nIndex([\'start\', \'end\', \'today\', \'username\', \'deviceid\', \'phonenumber\',\n       \'Date of the survey\', \'Health zone\', \'Name of the village\', \'team number\',\n       \'Cluster number\', \'Household number\', \'GPS Coordinates\',\n       \'_GPS Coordinates_latitude\', \'_GPS Coordinates_longitude\',\n       \'_GPS Coordinates_altitude\', \'_GPS Coordinates_precision\',\n       \'Head of the household present?\',\n       \'Age of the head of the household \',\n       \'Consent obtained\',\n       \'Number of children in the household\',\n       \'__version__\', \'_id\', \'_uuid\', \'_status\', \'_submission_time\', \'_tags\',\n       \'_notes\', \'_validation_status\', \'_submitted_by\', \'_index\'],\n      dtype=\'object\')\n\n\n# You can go back and forth between names and labels as much as you want \nmy_form.display(columns_as=\'label\', choices_as=\'name\')\nmy_form.display(columns_as=\'name\', choices_as=\'label\')\nmy_form.display(columns_as=\'name\', choices_as=\'name\')\nmy_form.display(columns_as=\'label\', choices_as=\'label\')\n\n```\n#### Note\n* For questions of type `select_multiple` the different answers are separated by a \'|\'.\n\n* If a form contains `n` columns with the same label, a suffix `(1)` to `(n)` will be added to each of the columns.\n\n### Repeats\n\n[Repeats](https://xlsform.org/en/#repeats) are supported (only one level, not repeats inside repeats).\nIn this case data of the repeat groups are separated from the \'main\' data and accessible via the \'repeats\' attribute\nwhich returns a Python dictionary\n\n```python\n\nprint(my_form.has_repeats)\n\nTrue\n# This means that the form has at least 1 repeat group\n\nprint(my_form.repeats.keys())\n\ndict_keys([\'children_questions\'])\n# The form has 1 repeat group called \'children_questions\'\n\n\nprint(my_form.repeats[\'children_questions\'])\n\n\n     index_repeat Sex of the child Age of the child  ... Going to school?  _parent_index\n0               1                           Male              No                    2\n1               2                           Female            No                    2\n2               1                           Female            No                    4\n3               1                           Female            Yes                   5\n4               2                           Female            No                    5\n...\n...\n[1040 rows x 27 columns]\n\n```\nThe column `_index` in the main DataFrame (my_form.data) and the column `_parent_index` in the DatFrame of the repeat\ngroup can be used to join the 2 DataFrames.\n\n```python\n\ndf_join = pd.merge(\n    my_form.data,\n    my_form.repeats[\'groupe_questions_enfants\'],\n    how="left",\n    left_on=\'_index\',\n    right_on=\'_parent_index\'\n)\n\n```\n### Save the data to file\n\nBecause the data is a pandas DataFrame, we can take advantage of the [many](https://pandas.pydata.org/docs/user_guide/io.html) pandas methods to export it to a file.\n\n```python\n# CSV\ndf_join.to_csv(\'household_survey.csv\', index=False)\n\n# Excel\ndf_join.to_excel(\'household_survey.xlsx\', index=False)\n```\n\n### Download a form in XLS or XML format\n\n```python\nmy_form.download_form(\'xls\')\n```\nThis downloads the XLSForm `tpz2buHAdXxcN0JVrZaSdk.xls` in the current working directory\n\n## Also\nPykobo has a bunch of utility methods that make easy to clean you data (not documented yet).\n\n## Note\nPykobo only reads and fetches data from Kobo forms. It doesn\'t update or delete the forms and their data on the Kobo server.\n\n## Dependencies\n* requests\n* pandas\n* numpy\n\n## TO DO\n* Add possibility to display group name as a prefix\n* Add method to download media files\n* Clean and document utility functions\n* Be more consistent and robust in case of errors\n* Calculate stats on forms time duration',
    'author': 'pvernier',
    'author_email': 'pvernier82@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import requests

from .form import KoboForm


class Manager:
    def __init__(self, url_api: str, token: str) -> None:
        self.url_api = url_api
        self.token = token
        self.headers = {"Authorization": f"Token {token}"}
        self._assets = None

    def _fetch_forms(self) -> None:
        """Fetch the list of forms the user has access to with its token."""
        url_assets = f"{self.url_api}/assets.json"

        res = requests.get(url=url_assets, headers=self.headers)

        # If error while fetching the data, return an empty list
        if res.status_code != 200:
            return []

        results = res.json()["results"]

        # It seems that when uploading an XLSForm from the website to create
        # a new form and there is an issue during the upload, the form
        # will be visible in the API but not in the UI. In this case it will
        # have the property "asset_type" set to "empty" instead of "survey"
        # for a working form. We don't want to keep them so we filter them out.
        # This issue seems to be very rare.
        results = [r for r in results if r["asset_type"] != "empty"]

        self._assets = results

    def _create_koboform(self, form: dict) -> KoboForm:
        kform = KoboForm(uid=form["uid"])
        kform._extract_from_asset(form)
        kform.headers = self.headers

        return kform

    def get_forms(self) -> list[KoboForm]:
        if not self._assets:
            self._fetch_forms()

        kforms = []
        for form in self._assets:
            kform = self._create_koboform(form)
            kforms.append(kform)
        return kforms

    def get_form(self, uid: str) -> KoboForm:
        if not self._assets:
            self._fetch_forms()

        form = [f for f in self._assets if f["uid"] == uid][0]
        kform = self._create_koboform(form)

        return kform



import pandas as pd


def clean_df(df: pd.DataFrame, columns: list) -> pd.DataFrame:
    """
    Given a Pandas DataFrame and a list of columns
    only keep the columns of the DataFrame that are in the list
    """

    return df[columns]


def lowercase_columns_name(df: pd.DataFrame) -> None:
    """
    Lowercase the name of all columns of the DataFrame 'df'
    """
    df.columns = df.columns.str.lower()


def lowercase_column_values(df: pd.DataFrame, column: str) -> None:
    """
    Lowercase the values of the column 'column' in the DataFramce 'df'
    """
    df[column] = df[column].str.lower()


def capitalize_column_values(df: pd.DataFrame, column: str) -> None:
    """
    Capitalize the values of the column 'column' in the DataFramce 'df'
    """
    df[column] = df[column].str.capitalize()


def trim_columns_values(df: pd.DataFrame) -> pd.DataFrame:
    """
    Trim whitespace from ends of each value across all series in dataframe
    Source: https://stackoverflow.com/questions/40950310/strip-trim-all-strings-of-a-dataframe
    """

    def trim_ws(x):
        return x.strip() if isinstance(x, str) else x

    def trim_pt(x):
        return x.strip(".") if isinstance(x, str) else x

    df_temp = df.applymap(trim_ws)
    return df_temp.applymap(trim_pt)


def reconcile_columns(
    df: pd.DataFrame, col1: str, col2: str, new_column: str, criteria=float("nan")
) -> None:
    """
    Given a DF, reconcile its 2 columns `col1` and `col2` into a new column `new_column`, as follow:
        if `col1` is null get the values from `col2` into `new_column`
        if `col1` is not null get the values from `col1` into `new_column`
    This means that `col1` is considered more important than `col2`
    Then delete columns `col1` and `col2`
    """

    # `_temp` is a temporary name for `new_column`

    if pd.isna(criteria):
        df.loc[(df[col1].isnull()), "_temp"] = df[col2]

        df.loc[(df[col1].notnull()), "_temp"] = df[col1]
    elif type(criteria) == str:
        df.loc[(df[col1] == criteria), "_temp"] = df[col2]

        df.loc[(df[col1] != criteria), "_temp"] = df[col1]
    else:
        raise TypeError(
            f"The argument `criteria` is of type: `{type(criteria)}`. It should be of type `<class 'str'>` or equal to `NaN`"
        )

    # Delete the 2 columns
    df.drop(columns=[col1, col2], inplace=True)

    # Rename column `_temp` to `new_column`
    df.rename(columns={"_temp": new_column}, inplace=True)


def reconcile_columns_append(
    df: pd.DataFrame, col1: str, col2: str, new_column: str, to_replace: str
) -> None:
    """TODO
    XXX: See if can merge this function and the one above into 1
    """

    # If the column is not of type str we convert it
    # This can happen for columns that are empty and were
    # not in the JSON object so we created empty column containing NaN
    if df[col1].dtype != "object":
        df[col1] = df[col1].astype(str)

    selection = df[col1].str.contains(to_replace, na=False)
    for index, row in df.loc[selection].iterrows():
        df.loc[index, col1] = row[col1].replace(to_replace, row[col2])

    df["_temp"] = df[col1]

    # Delete the 2 columns
    df.drop(columns=[col1, col2], inplace=True)

    # Rename column `_temp` to `new_column`
    df.rename(columns={"_temp": new_column}, inplace=True)


def fix_typos(df: pd.DataFrame, column: str, typos: list) -> None:
    for typo in typos:
        df.loc[df[column] == typo[0], column] = typo[1]


def fillna(df: pd.DataFrame, column: str) -> None:
    df[column] = df[column].fillna(0)


def convert_toint(df: pd.DataFrame, column) -> None:
    df[column] = df[column].astype(int)
